from copy import copy
from logging import Logger
from logging import LogRecord
from logging import getLogger

import click
from typing_extensions import Self
from uvicorn.logging import AccessFormatter
from uvicorn.logging import DefaultFormatter

logger: Logger = getLogger(__name__)

TIMEZONE_FORMAT = "%Y-%m-%d %H:%M:%S"


class TimedAccessFormatter(AccessFormatter):
    default_time_format = TIMEZONE_FORMAT

    @staticmethod
    def get_color_time(time: float) -> str:
        s: str = f"{time:.6f}"

        if time < 0.5:
            s = click.style(str(s), fg="green")
        elif time < 0.7:
            s = click.style(str(s), fg="yellow")
        elif time < 1.0:
            s = click.style(str(s), fg="red")
        else:
            s = click.style(str(s), fg="bright_red")

        return s

    def formatMessage(self: Self, record: LogRecord) -> str:  # noqa: N802
        recordcopy = copy(record)

        if len(record.args) != 7:
            return super().formatMessage(recordcopy)

        (
            client_addr,
            method,
            full_path,
            http_version,
            status_code,
            time,
            response_length,
        ) = recordcopy.args  # type: ignore[misc]

        time = self.get_color_time(time) if self.use_colors else f"{time:.6f}"

        recordcopy.__dict__.update(
            {
                "run_time": time,
                "response_length": response_length,
            }
        )

        recordcopy.args = (
            client_addr,
            method,
            full_path,
            http_version,
            status_code,
        )

        return super().formatMessage(recordcopy)


class ConsoleFormatter(DefaultFormatter):
    default_time_format = TIMEZONE_FORMAT



__version__ = "0.0.7"  # type: str

from logging import Logger
from logging import getLogger

from .application import FastAPI
from .responses import error_response
from .responses import success_response

logger: Logger = getLogger(__name__)



from logging import Logger
from logging import getLogger
from typing import TYPE_CHECKING
from typing import Any

import httpx
import sentry_sdk
from fastapi import FastAPI
from httpx._urlparse import ParseResult
from httpx._urlparse import urlparse
from sentry_sdk.client import _get_options

logger: Logger = getLogger(__name__)


def _init(*args: str | None, **kwargs: Any) -> None:
    """
    Initialize Sentry SDK, add the url check before initializing.
    Read sentry_sdk.init for more information.

    Args:
        *args:
        **kwargs:

    Returns:
        None
    """
    if len(args) >= 1 and issubclass(type(args[0]), FastAPI):
        args = args[1:]

    if args and isinstance(args[0], str):
        dsn: str = args[0]  # DSN is the first argument
    else:
        dsn: str = kwargs.get("dsn") or _get_options().get("dsn")

    if dsn is None:
        logger.debug("DSN is not provided.")
        return

    if dsn == "":
        logger.debug("DSN is empty.")
        return

    parse_url: ParseResult = urlparse(dsn)
    test_connection: httpx.Response = httpx.get(f"{parse_url.scheme}://{parse_url.netloc}")

    if test_connection.is_success or test_connection.is_redirect:
        sentry_sdk.init(*args, **kwargs)

        logger.info("Sentry enabled.")


# Fake sentry_init for type hinting
if TYPE_CHECKING:
    from sentry_sdk.consts import ClientConstructor
    from sentry_sdk.hub import _InitGuard

    class sentry_init(ClientConstructor, _InitGuard):  # noqa: N801
        pass
else:
    sentry_init = (lambda: _init)()



import math
import os
from collections.abc import Generator
from pathlib import Path
from unittest import mock

import pytest
from typing_extensions import Self

from api_helper.config import is_debug
from api_helper.config import is_reload
from api_helper.config.configure import _auto_convert
from api_helper.config.configure import _convert_bool
from api_helper.config.configure import get
from api_helper.config.configure import load_config
from api_helper.errors import ConfigError


@pytest.fixture()
def set_debug_env(monkeypatch: pytest.MonkeyPatch) -> Generator[None, None, None]:
    with mock.patch.dict(os.environ, clear=True):
        envvars = {
            "DEBUG": "true",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)
        yield  # This is the magical bit which restore the environment after


@pytest.fixture()
def set_reload_env(monkeypatch: pytest.MonkeyPatch) -> Generator[None, None, None]:
    with mock.patch.dict(os.environ, clear=True):
        envvars = {
            "RELOAD": "true",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)
        yield  # This is the magical bit which restore the environment after


@pytest.fixture()
def set_test_envs(monkeypatch: pytest.MonkeyPatch) -> Generator[None, None, None]:
    with mock.patch.dict(os.environ, clear=True):
        envvars = {
            "TEST_TRUE": "true",
            "TEST_FALSE": "false",
            "TEST_INT": "255",
            "TEST_FLOAT": "3.14",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)
        yield  # This is the magical bit which restore the environment after


class TestDebug:
    def test_is_debug_default(self: Self) -> None:
        assert is_debug() is False

    def test_is_debug_set_true(self: Self, set_debug_env: None) -> None:
        assert is_debug() is True


class TestReload:
    def test_is_reload_default(self: Self) -> None:
        assert is_reload() is False

    def test_is_reload_set_true(self: Self, set_reload_env: None) -> None:
        assert is_reload() is True


class TestConvert:
    def test_convert_bool(self: Self) -> None:
        for value in ["True", "true", "1"]:
            assert _convert_bool(value) is True

        for value in ["False", "false", "0"]:
            assert _convert_bool(value) is False

    def test_convert_bool_error(self: Self) -> None:
        with pytest.raises(ConfigError) as error:
            _convert_bool("invalid")

        assert error.match("^Cannot convert.*$")


class TestAutoConvert:
    def test_convert_none(self: Self) -> None:
        assert _auto_convert(None) is None

    def test_convert_bool(self: Self) -> None:
        assert _auto_convert("true") is True
        assert _auto_convert("false") is False
        assert _auto_convert("1") is True
        assert _auto_convert("0") is False
        assert _auto_convert("yes") is True
        assert _auto_convert("no") is False

    def test_convert_int(self: Self) -> None:
        assert _auto_convert("1") == 1
        assert _auto_convert("0") == 0
        assert _auto_convert("-1") == -1

    def test_convert_float(self: Self) -> None:
        assert math.isclose(_auto_convert("1.0"), 1.0, rel_tol=1e-09, abs_tol=1e-09)
        assert math.isclose(_auto_convert("0.0"), 0.0, rel_tol=1e-09, abs_tol=1e-09)
        assert math.isclose(_auto_convert("-1.0"), -1.0, rel_tol=1e-09, abs_tol=1e-09)

    def test_convert_list(self: Self) -> None:
        assert _auto_convert("[1, 2, 3]") == [1, 2, 3]
        assert _auto_convert('["a", "b", "c"]') == ["a", "b", "c"]
        assert _auto_convert("[1, 2.0, 'c']") == [1, 2.0, "c"]

    def test_convert_dict(self: Self) -> None:
        assert _auto_convert('{"a": 1, "b": 2, "c": 3}') == {"a": 1, "b": 2, "c": 3}
        assert _auto_convert('{"a": "a", "b": "b", "c": "c"}') == {"a": "a", "b": "b", "c": "c"}
        assert _auto_convert('{"a": 1, "b": 2.0, "c": "c"}') == {"a": 1, "b": 2.0, "c": "c"}

    def test_convert_string(self: Self) -> None:
        assert _auto_convert("a") == "a"
        assert _auto_convert("b") == "b"
        assert _auto_convert("long_long_string") == "long_long_string"

    def test_convert_object(self: Self) -> None:
        with pytest.raises(ConfigError) as error:
            _auto_convert(object)

        assert error.match("^Cannot convert.*, should be a string or None.$")


class TestGet:
    def test_get_default_key(self: Self) -> None:
        assert get("key") is None
        assert get("key", None) is None
        assert get("key", "default") == "default"

    def test_get_convert(self: Self, set_test_envs: None) -> None:
        assert get("TEST_TRUE", convert=True) is True
        assert get("TEST_FALSE", convert=True) is False
        assert get("TEST_INT", convert=True) == 255
        assert math.isclose(get("TEST_FLOAT", convert=True), 3.14, rel_tol=1e-09, abs_tol=1e-09)

    def test_get_raise_error(self: Self) -> None:
        with pytest.raises(ConfigError) as error:
            get("key", raise_error=True)

        assert error.match("^Key 'key' not found.$")


class TestLoadConfig:
    def test_load_config(self: Self) -> None:
        result = load_config(Path(__file__).parent.parent / "statics/sample_env")

        assert result is None

    def test_load_config_dir(self: Self) -> None:
        Path.touch(Path(__file__).parent.parent / "statics/.env")

        result = load_config(Path(__file__).parent.parent / "statics")

        assert result is None

    def test_load_not_exist_config(self: Self) -> None:
        with pytest.raises(ConfigError) as error:
            load_config(Path("not_exist"))

        assert error.match("^The env file not_exist does not exist or is not a directory containing .env file.$")



import logging.config
from logging import Logger
from logging import getLogger
from pathlib import Path
from typing import Any

from typing_extensions import Self

from .. import config
from .formatter import ConsoleFormatter
from .formatter import TimedAccessFormatter

logger: Logger = getLogger(__name__)

FORMATS: dict[str, str] = {
    "default": "[%(levelname)s] %(asctime)s - %(name)s - %(message)s",
    "access": (
        "[%(levelname)s] %(asctime)s %(client_addr)s - %(run_time)s "
        '"%(request_line)s" %(status_code)s %(response_length)s'
    ),
}
FORMATTER: dict[str, type[logging.Formatter]] = {
    "console": ConsoleFormatter,
    "access": TimedAccessFormatter,
}
FORMATTERS: dict[str, Any] = {
    "console": {
        "()": FORMATTER["console"],
        "fmt": FORMATS["default"],
        "use_colors": True,
    },
    "file": {
        "()": FORMATTER["console"],
        "fmt": FORMATS["default"],
    },
    "access": {
        "()": FORMATTER["access"],
        "fmt": FORMATS["access"],
        "use_colors": True,
    },
    "access_file": {
        "()": FORMATTER["access"],
        "fmt": FORMATS["access"],
        "use_colors": False,
    },
}
HANDLER: dict[str, str] = {
    "stream": "logging.StreamHandler",
    "file": "logging.handlers.TimedRotatingFileHandler",
}


class _LoggerBuilder:
    __slots__ = ("base_folder", "log_folder", "log_files", "formatters", "handlers", "loggers", "config")

    def __init__(self: Self, base_folder: Path, folder_name: str = "logs", **kwargs: Any) -> None:
        self.base_folder: Path = base_folder

        self.log_folder: Path = self.base_folder / folder_name
        self.log_files: dict[str, Path] = {
            "console": self.log_folder / "console.log",
            "access": self.log_folder / "access.log",
        }

        self.formatters: dict[str, Any] = kwargs.get("formatters") or FORMATTERS
        self.handlers: dict[str, Any] = kwargs.get("handlers") or {
            "stdout": {
                "formatter": "console",
                "class": HANDLER["stream"],
                "stream": "ext://sys.stdout",
            },
            "stderr": {
                "formatter": "console",
                "class": HANDLER["stream"],
                "stream": "ext://sys.stderr",
            },
            "consolelog": {
                "formatter": "file",
                "class": HANDLER["file"],
                "filename": self.log_files["console"],
                "when": "D",
                "delay": True,
                "backupCount": 5,
                "encoding": "utf8",
            },
            "access": {"formatter": "access", "class": HANDLER["stream"], "stream": "ext://sys.stdout"},
            "access_file": {
                "formatter": "access_file",
                "class": HANDLER["file"],
                "filename": self.log_files["access"],
                "when": "D",
                "delay": True,
                "backupCount": 5,
                "encoding": "utf8",
            },
        }
        self.loggers: dict[str, Any] = kwargs.get("loggers") or {
            "api_helper": {"level": "ERROR"},
            "service.error": {"handlers": ["stderr", "consolelog"], "level": "ERROR", "propagate": False},
            "access_time": {"handlers": ["access", "consolelog", "access_file"], "level": "INFO", "propagate": False},
            "uvicorn": {"handlers": ["stderr", "consolelog"], "level": "INFO", "propagate": False},
            "uvicorn.error": {"level": "INFO"},
            "uvicorn.access": {"level": "INFO", "propagate": False},
            "sentry_sdk": {"level": "ERROR"},
            "urllib3": {"level": "INFO"},
            "httpx": {"level": "INFO"},
            "httpcore": {"level": "INFO"},
            "watchfiles": {"level": "INFO"},
        }
        self.config: dict[str, Any] = kwargs.get("config") or {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": self.formatters,
            "handlers": self.handlers,
            "loggers": self.loggers,
            "root": {"handlers": ["stdout", "consolelog"], "level": "DEBUG" if config.is_debug() else "INFO"},
        }

    def setup(self: Self) -> None:
        if not Path.exists(self.log_folder):
            Path.mkdir(self.log_folder, parents=True)

        # For logfiles, input new line to indicate the start of the log
        for logfile in self.log_files.values():
            with Path.open(logfile, "a") as f:
                f.write("====== NEW APPLICATION START ======\n")

        root_logger: Logger = logging.root
        root_logger.setLevel(logging.INFO)

        # It means the logging config is not set, using the default config
        if len(root_logger.handlers) == 0:
            logging.config.dictConfig(self.config)
            logger.debug("Using default logging config.")



import time
import traceback
import typing as t
from http import HTTPStatus
from logging import Logger
from logging import getLogger

from starlette.background import BackgroundTask
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp
from typing_extensions import Self
from uvicorn.protocols.utils import get_client_addr
from uvicorn.protocols.utils import get_path_with_query_string

from . import config
from .responses import error_response

logger: Logger = getLogger(__name__)

UNKNOWN_ERROR_CODE: int = HTTPStatus.INTERNAL_SERVER_ERROR
UNKNOWN_ERROR_DESCRIPTION: str = "An unknown internal error has occurred while processing the request."


class LogRequestMiddleware(BaseHTTPMiddleware):
    def __init__(
        self: Self,
        app: ASGIApp,
    ) -> None:
        super().__init__(app)

    async def dispatch(self: Self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        # process the request and get the response
        start_time: float = time.time()

        try:
            response: Response = await call_next(request)

        except Exception as e:  # noqa: BLE001
            content: dict[str, t.Any] = {
                "code": UNKNOWN_ERROR_CODE,
                "detail": UNKNOWN_ERROR_DESCRIPTION,
                "parameters": {
                    "arguments": e.args,
                },
            }

            if config.is_debug():
                content["parameters"]["traceback"] = traceback.format_exc().split("\n")

            response = error_response(
                data=content,
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
            )

            self.log_error(request, response)

        process_time: float = time.time() - start_time

        response.headers["X-Process-Time"] = str(process_time)
        response.background = BackgroundTask(
            self.log_time,
            request,
            response,
            process_time,
        )

        return response

    @staticmethod
    def log_error(request: Request, response: Response) -> None:
        service_logger: Logger = getLogger("service.error")

        service_logger.info(
            '%s - "%s %s HTTP/%s" %d',
            get_client_addr(request.scope),  # type: ignore
            request.scope["method"],
            get_path_with_query_string(request.scope),  # type: ignore
            request.scope["http_version"],
            response.status_code,
        )
        service_logger.error(
            "Error in ASGI application",
            exc_info=True,
        )

    @staticmethod
    def log_time(request: Request, response: Response, _time: float) -> None:
        access_log: Logger = getLogger("access_time")

        access_log.info(
            '%s - "%s %s HTTP/%s" %d, response time: %f, response bytes: %s',
            get_client_addr(request.scope),  # type: ignore
            request.scope["method"],
            get_path_with_query_string(request.scope),  # type: ignore
            request.scope["http_version"],
            response.status_code,
            _time,
            response.headers.get("Content-Length", 0),
        )



import logging
from pathlib import Path

import httpx
import pytest
from typing_extensions import Self

from api_helper import FastAPI
from api_helper.errors import FastAPIError


class MockHttpGet:
    @staticmethod
    def is_success() -> bool:
        return True


@pytest.fixture()
def success_httpx(monkeypatch: pytest.MonkeyPatch) -> None:
    def mock_get(*args, **kwargs) -> MockHttpGet:
        return MockHttpGet()

    monkeypatch.setattr(httpx, "get", mock_get)


class TestApplication:
    def test_fastapi_basefolder_not_exist(self: Self) -> None:
        with pytest.raises(FastAPIError) as error:
            FastAPI(base_folder=Path("not_exist"))

        assert error.match("^The base folder does not exist.*$")

    def test_fastapi(self: Self, tmp_path: Path) -> None:
        app: FastAPI = FastAPI(base_folder=tmp_path)

        assert app

    def test_fastapi_sentry_no_sentry_config(self: Self, tmp_path: Path, caplog: pytest.LogCaptureFixture) -> None:
        app: FastAPI = FastAPI(base_folder=tmp_path)

        caplog.set_level(logging.DEBUG)
        app.setup_sentry()

        assert "DSN is not provided." in caplog.text

    def test_fastapi_sentry_empty_str(self: Self, tmp_path: Path, caplog: pytest.LogCaptureFixture) -> None:
        app: FastAPI = FastAPI(base_folder=tmp_path)

        caplog.set_level(logging.DEBUG)
        app.setup_sentry("")

        assert "DSN is empty." in caplog.text

    def test_fastapi_sentry_valid_arg_config(
        self: Self, tmp_path: Path, caplog: pytest.LogCaptureFixture, success_httpx: None
    ) -> None:
        app: FastAPI = FastAPI(base_folder=tmp_path)
        app.setup_sentry("http://samplecode@example.com/10")

        assert "Sentry enabled." in caplog.text

    def test_fastapi_sentry_valid_kwarg_config(
        self: Self, tmp_path: Path, caplog: pytest.LogCaptureFixture, success_httpx: None
    ) -> None:
        app: FastAPI = FastAPI(base_folder=tmp_path)
        app.setup_sentry(dsn="http://samplecode@example.com/10")

        assert "Sentry enabled." in caplog.text



import contextlib
import logging.config
import os
from ast import literal_eval
from logging import Logger
from pathlib import Path
from typing import Any

from dotenv import load_dotenv

from ..errors import ConfigError

logger: Logger = logging.getLogger(__name__)

FALLBACK_CONFIGS: dict[str, Any] = {
    "APP_NAME": "Backend API",
    "DEBUG": False,
    "RELOAD": False,
}


def _convert_bool(value: str) -> bool:
    """Convert the string value to boolean value.

    Args:
    ----
        value: The string value

    Returns:
    -------
        bool: The boolean value

    """
    if value.lower() in ["true", "t", "1", "yes", "y", "on"]:
        return True

    elif value.lower() in ["false", "f", "0", "no", "n", "off"]:
        return False

    else:
        msg: str = f"Cannot convert {value} to boolean value."
        raise ConfigError(msg)


def _auto_convert(env: Any) -> Any:
    if isinstance(env, str):
        try:
            env = _convert_bool(env)

        except ConfigError:
            with contextlib.suppress(ValueError, SyntaxError):
                # Auto guess type, only support int, float, list, dict
                env = literal_eval(env)

    elif env is None:
        env = None

    else:
        msg: str = f"Cannot convert {env}, should be a string or None."
        raise ConfigError(msg)

    return env


def get(key: str, default: Any = None, *, convert: bool = False, raise_error: bool = False) -> Any:
    """Get the value of the key from the config file, if not found, return the default value.

    Args:
    ----
        key: The key of the config
        default: The default value if the key is not found
        convert: If the value should be converted to the correct type, otherwise, return the string type value
        raise_error: If the key is not found and the default value is None, raise the KeyError

    Returns:
    -------
        Any: The value of the key

    """
    logger.debug("Get config key %s", key)
    env: Any = os.getenv(key)

    if convert:
        env = _auto_convert(env)

    if env is not None:
        return env

    if key in FALLBACK_CONFIGS and FALLBACK_CONFIGS[key] is not None:
        return FALLBACK_CONFIGS[key]

    else:
        if raise_error and default is None:
            msg: str = f"Key '{key}' not found."
            raise ConfigError(msg)

        return default


def load_config(env_file: Path) -> None:
    logger.debug("Load env file from %s", env_file)

    if env_file.is_file() and env_file.exists():
        load_dotenv(dotenv_path=env_file, override=True)

    elif env_file.is_dir() and (dot_env := env_file / ".env").exists():
        load_dotenv(dotenv_path=dot_env, override=True)

    else:
        msg: str = f"The env file {env_file} does not exist or is not a directory containing .env file."
        raise ConfigError(msg)



import logging
from pathlib import Path

from pytest_mock import MockFixture
from typing_extensions import Self

from api_helper.config import TimedAccessFormatter
from api_helper.config import _LoggerBuilder


class TestFastAPIConfigWrap:
    def test_config(self: Self, tmp_path: Path) -> None:
        config: _LoggerBuilder = _LoggerBuilder(tmp_path)

        assert config.base_folder == tmp_path
        assert config.log_folder == tmp_path / "logs"

    def test_log_setup(self: Self, tmp_path: Path) -> None:
        config: _LoggerBuilder = _LoggerBuilder(tmp_path)
        config.setup()

    def test_log_mock_root_handler(self: Self, mocker: MockFixture, tmp_path: Path) -> None:
        logger_mock: logging.Logger = mocker.patch.object(logging, "root", logging.getLogger("_tmp_"))

        config: _LoggerBuilder = _LoggerBuilder(tmp_path)
        config.setup()

        assert logger_mock


class TestFastAPIFormatter:
    def test_timed_access_formatter(self: Self) -> None:
        formatter: TimedAccessFormatter = TimedAccessFormatter()
        result_03: str = formatter.get_color_time(0.3)
        result_05: str = formatter.get_color_time(0.5)
        result_07: str = formatter.get_color_time(0.7)
        result_10: str = formatter.get_color_time(1.0)
        result_11: str = formatter.get_color_time(1.1)

        assert result_03 == "\x1b[32m0.300000\x1b[0m"
        assert result_05 == "\x1b[33m0.500000\x1b[0m"
        assert result_07 == "\x1b[31m0.700000\x1b[0m"
        assert result_10 == "\x1b[91m1.000000\x1b[0m"
        assert result_11 == "\x1b[91m1.100000\x1b[0m"

    def test_message(self: Self) -> None:
        formatter: TimedAccessFormatter = TimedAccessFormatter()

        record: logging.LogRecord = logging.LogRecord(
            name="test",
            level=20,
            pathname="test",
            lineno=1,
            msg="test",
            args=("localhost", "GET", "/test", "HTTP/1.1", 200, 0.5, 0),
            exc_info=None,
        )
        record.message = "test"
        message: str = formatter.formatMessage(record)
        assert message == "test"

        record: logging.LogRecord = logging.LogRecord(
            name="test",
            level=20,
            pathname="test",
            lineno=1,
            msg="test",
            args=("localhost", "GET", "/test", "HTTP/1.1", 200),
            exc_info=None,
        )
        record.message = "5word"
        message: str = formatter.formatMessage(record)
        assert message == "5word"



import math
import typing as t
from http import HTTPStatus
from logging import Logger
from logging import getLogger

import arrow
import orjson
from pydantic import BaseModel
from starlette.background import BackgroundTask
from starlette.responses import JSONResponse
from starlette.responses import Response
from typing_extensions import Self

logger: Logger = getLogger(__name__)


try:
    from sqlalchemy.orm import DeclarativeMeta

    SQLALCHEMY_INSTALLED: bool = True

except ImportError:
    SQLALCHEMY_INSTALLED: bool = False


def _default(obj: t.Any) -> t.Any:
    if SQLALCHEMY_INSTALLED and isinstance(type(obj), DeclarativeMeta):
        # Return obj.__dict__ if it exists and remove the _sa_instance_state
        return {key: value for key, value in obj.__dict__.items() if not key.startswith("_")}

    if isinstance(obj, BaseModel):
        return obj.model_dump()

    raise TypeError


class ORJSONResponse(JSONResponse):
    # Override the default JSONResponse to use orjson
    # same from fastapi.responses.ORJSONResponse

    def __init__(
        self: Self,
        content: t.Any,
        status_code: int = 200,
        headers: t.Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None,
        # Customized default to use orjson
        default: t.Callable[[t.Any], t.Any] = _default,
    ) -> None:
        self.default = default

        super().__init__(content, status_code, headers, media_type, background)

    def render(self: Self, content: t.Any) -> bytes:
        return orjson.dumps(content, option=orjson.OPT_NON_STR_KEYS | orjson.OPT_SERIALIZE_NUMPY, default=self.default)


class SuccessResponse(ORJSONResponse):
    key: str = "data"

    def __init__(
        self: Self,
        data: t.Iterable[bytes] | bytes | t.Iterable[str] | str | None = None,
        *,
        status_code: int | HTTPStatus = HTTPStatus.OK,
        headers: t.Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        super().__init__(
            self.generate_content(data),
            status_code,
            headers,
            media_type,
            background,
        )

    def generate_content(self: Self, data: t.Any) -> dict[str, t.Any]:
        if not data:
            data = ""

        return {
            self.key: data,
            "datetime": arrow.utcnow().isoformat(),
        }


class ErrorResponse(SuccessResponse):
    key: str = "error"


def success_response(
    data: t.Iterable[bytes] | bytes | t.Iterable[str] | str | BaseModel | None = None,
    *,
    status_code: int | HTTPStatus = HTTPStatus.OK,
    headers: t.Mapping[str, str] | None = None,
    media_type: str | None = None,
    background: BackgroundTask | None = None,
) -> Response:
    if math.floor(status_code / 100) in (4, 5):
        msg: str = "4xx and 5xx are not success response."
        raise ValueError(msg)

    if not data:
        data = "success"

    return SuccessResponse(
        data,
        status_code=status_code,
        headers=headers,
        media_type=media_type,
        background=background,
    )


def error_response(
    data: t.Iterable[bytes] | bytes | t.Iterable[str] | str | None = None,
    *,
    status_code: int | HTTPStatus = HTTPStatus.BAD_REQUEST,
    headers: t.Mapping[str, str] | None = None,
    media_type: str | None = None,
    background: BackgroundTask | None = None,
) -> Response:
    if math.floor(status_code / 100) not in (4, 5):
        msg: str = "4xx and 5xx are not error response."
        raise ValueError(msg)

    if not data:
        data = "error"

    return ErrorResponse(
        data,
        status_code=status_code,
        headers=headers,
        media_type=media_type,
        background=background,
    )



from .configure import get
from .configure import load_config
from .formatter import ConsoleFormatter
from .formatter import TimedAccessFormatter
from .logger import _LoggerBuilder


def is_debug() -> bool:
    return get("DEBUG", False, convert=True)


def is_reload() -> bool:
    return get("RELOAD", False, convert=True)



from collections.abc import Callable
from logging import Logger
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import uvicorn
from fastapi import FastAPI as FastAPIBase
from starlette.middleware import Middleware
from typing_extensions import Self
from uvicorn._types import ASGIApplication

from . import config
from .config import _LoggerBuilder
from .errors import FastAPIError
from .middlewares import LogRequestMiddleware
from .sentry import sentry_init

logger: Logger = getLogger(__name__)


class FastAPI(FastAPIBase):
    def __init__(
        self: Self,
        *,
        base_folder: Path,
        **extra: Any,
    ) -> None:
        """

        Args:
            base_folder:
            **extra:
        """
        if not base_folder.exists():
            msg: str = "The base folder does not exist. Please provide a valid base folder."
            raise FastAPIError(msg)

        options: dict[str, Any] = {}
        for key in extra:
            if key.startswith("logging_"):
                options[key.replace("logging_", "")] = extra.pop(key)

        _logger: _LoggerBuilder = _LoggerBuilder(base_folder, **options)
        _logger.setup()

        self.LOGGING_CONFIG: dict[str, Any] = _logger.config

        middlewares: list[Middleware] = extra.pop("middleware", [])
        middlewares.append(
            Middleware(LogRequestMiddleware),  # type: ignore
        )

        if config.is_debug() or extra.get("debug", False) is True:
            super().__init__(
                debug=True,
                title=config.get("APP_NAME"),
                middleware=middlewares,
                **extra,
            )
        else:
            openapi_url: str | None = extra.pop("openapi_url", None)

            super().__init__(
                title=config.get("APP_NAME"),
                middleware=middlewares,
                openapi_url=openapi_url,
                **extra,
            )

    if TYPE_CHECKING:
        run = uvicorn.run

    else:

        def run(
            self: Self,
            host: str = "127.0.0.1",
            port: int = 8000,
            *,
            show_swagger: bool = False,
            app: ASGIApplication | Callable[..., Any] | str | None = None,
            **kwargs: Any,
        ) -> None:  # pragma: no cover
            error_message: str = ""

            if show_swagger and self.openapi_url is None:
                if self.title is None:
                    error_message = "A title must be provided for OpenAPI, e.g.: 'My API'"
                if self.version is None:
                    error_message = "A version must be provided for OpenAPI, e.g.: '2.1.0'"
                self.openapi_url = "/openapi.json"

                if error_message:
                    raise FastAPIError(error_message)

                self.setup()

            uvicorn.run(
                app or self,
                host=host,
                port=port,
                log_config=self.LOGGING_CONFIG,
                **kwargs,
            )

    setup_sentry = sentry_init
